def register_model() -> None:
  import json
  from vllm.logger import init_logger
  from vllm.model_executor.models.registry import ModelRegistry

  logger = init_logger(__name__)

  import pathlib
  _models_json = pathlib.Path(__file__).parent / 'models.json'

  with open(_models_json) as f:
    models: dict = json.load(f)

  for model_name in models.keys():
    ablated_model_name = f"Ablated{model_name}"
    ModelRegistry.register_model(
        ablated_model_name,
        f"ablated_vllm.model:{ablated_model_name}",
    )
    logger.info(f"Registered ablated_vllm.model:{ablated_model_name} with vLLM")
    print(f"Registered ablated_vllm.model:{ablated_model_name} with vLLM")
