from setuptools import find_packages, setup
# USE JSON TO LOAD WHICH MODELS TO PATCH


setup(
    name="ablated-vllm-plugin",
    version="0.0.1",
    description="Ablation plugin for vLLM",
    packages=find_packages(),
    package_data={"ablated_vllm": ["models.json"]},
    python_requires=">=3.10",
    entry_points={
        "vllm.general_plugins": [
            "ablated_vllm_models=ablated_vllm:register_model",
        ],
    },
)
